#!/usr/bin/env python

"""
Convert simple font bitmaps to GNU assembler data.

Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys

class MissingRow(Exception):

    "Indicate a missing row in a character definition."

    def __init__(self, lineno):
        self.lineno = lineno

    def __repr__(self):
        return "MissingRow(%r)" % self.lineno



def get_character_row(f, lineno=0):

    "Return a row of characters from 'f', employing 'lineno' for reference."

    line = f.readline().strip()
    rows = []

    while line:
        row = line.split()
        if rows and len(row) != len(rows[-1]):
            raise MissingRow, lineno

        rows.append(row)
        lineno += 1
        line = f.readline().strip()

    chars = []

    if rows:
        for i in range(0, len(rows[0])):
            char = []
            for row in rows:
                char.append(row[i])
            chars.append(char)

    return chars



def convert(infile, outfile, base=33):

    "Convert textual data in 'infile' to assembler data in 'outfile'."

    f_in = open(infile)
    f_out = open(outfile, "w")
    try:
        print >>f_out, """\
.globl fontchars
.globl fonttable
.globl fontbase
.globl fontlimit

.section .flash, "a"

fontbase:
.word %d

fontchars:""" % base

        lineno = 1
        offset = 0
        offsets = []
        point = base

        # Get each row of characters.

        chars = get_character_row(f_in, lineno)

        while chars:
            lineno += len(chars[0]) + 1

            # Write each character definition.

            for point, char in enumerate(chars, point):
                offsets.append(offset)
                offset += write_character(f_out, point, char)

            point += 1

            # Get the next row.

            chars = get_character_row(f_in, lineno)

        # Write the limit of characters.

        print >>f_out, """\
fontlimit:
.word %d""" % point

        # Write the offset table.

        print >>f_out, """\
fonttable:"""

        for offset in offsets:
            print >>f_out, ".word %d" % offset

    finally:
        f_in.close()
        f_out.close()



def write_character(f, point, char):

    "Write to 'f' for the given 'point' the indicated 'char' definition."

    # Write a comment and pixel width information.

    print >>f, """\
/* %d */
.byte %d /* width in pixels */
.byte %d /* height in pixels */
""" % (point, len(char[0]), len(char))

    # Write the pixel rows to a sequence of bytes.

    for row in char:
        print >>f, ".byte 0b%s" % row.replace("-", "0").replace("#", "1")

    print >>f

    # Return the size of the definition as the number of rows plus the width
    # and height bytes.

    return len(char) + 2



# Main program.

if __name__ == "__main__":

    # Obtain filenames.

    try:
        infile, outfile = sys.argv[1:3]
    except ValueError:
        print >>sys.stderr, """\
Usage: %s <input filename> <output filename>
"""
        sys.exit(1)

    # Perform the conversion.

    try:
        convert(infile, outfile)
    except MissingRow, exc:
        print >>sys.stderr, """\
Missing character information at line %d in file %s.
""" % (exc.lineno, infile)

# vim: tabstop=4 expandtab shiftwidth=4
